# -*- coding: utf-8 -*-

from simmate.toolkit import Composition, Structure
from simmate.toolkit.creators.structure.prototypes import FromAflowPrototypes


def get_structures_from_prototypes(
    composition: Composition,
    **kwargs,
) -> list[Structure]:
    """
    Generates a list of structures by using prototypes as templates. This function
    is a simple wrapper around the functionality in
    `simmate.toolkit.creators.structures.prototypes`. For more control over the
    structures generated, refer to the underlying creator class.
    """

    from_prototypes = FromAflowPrototypes(
        composition,
        **kwargs,
    ).structures

    return from_prototypes



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
)
from simmate.calculators.vasp.workflows.base import VaspWorkflow


class Relaxation__Vasp__Quality00(VaspWorkflow):
    """
    Runs a very rough VASP geometry optimization with fixed lattice volume.
    `Quality 00` indicates these are absolute lowest quality settings used in
    our available presets.

    Typically, you'd only want to run this relaxation on structures that were
    randomly created (and thus likely very unreasonable). More precise relaxations
    should be done afterwards. Therefore, instead of using this calculation,
    we recommend only using the relaxation/staged workflow, which uses this
    calculation as a first step.
    """

    # This uses the PBE functional with POTCARs that have lower electron counts
    # and convergence criteria when available.
    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS_LOW_QUALITY

    # because this calculation is such a low quality we don't raise an error
    # if the calculation fails to converge
    confirm_convergence = False

    # Make the unitcell relatively cubic before relaxing
    pre_sanitize_structure = True

    # These are all input settings for this task.
    incar = dict(
        # These settings are the same for all structures regardless of composition.
        PREC="Low",
        EDIFF=2e-3,
        EDIFFG=-2e-1,
        ISIF=4,  # this fixes lattice volume
        NSW=75,
        IBRION=2,  # for cases of bad starting sites
        POTIM=0.02,
        LCHARG=False,
        LWAVE=False,
        KSPACING=0.75,
        # The type of smearing we use depends on if we have a metal, semiconductor,
        # or insulator. So we need to decide this using a keyword modifier.
        multiple_keywords__smart_ismear={
            "metal": dict(
                ISMEAR=1,
                SIGMA=0.1,
            ),
            "non-metal": dict(
                ISMEAR=0,
                SIGMA=0.05,
            ),
        },
    )



# -*- coding: utf-8 -*-

import logging
from abc import ABC, abstractmethod


class Transformation(ABC):

    """
    This is an abstract base class (ABC) that shows you how to write your
    own Transformation class. It defines the requirements for you and raises
    errors if you forgot to define something.

    Using this abstract class, here's an example subclass that you could write
    -------------------------------------
    class SubClass(Transformation):
        io_type = 'one_to_one'
        ninput = 1
        use_multiprocessing = False
        def apply_transformation(self, structure):
            # do some transform on the input
            return new_structure
    -------------------------------------
    # example use
    t = SubClass()
    new_structure = t.apply_transformation(structure)
    -------------------------------------

    This is an updated version of pymatgen's transformation class because they
    do not support many_to_one or many_to_many:
    pymatgen.transformations.transformation_abc
    see https://pymatgen.org/pymatgen.transformations.transformation_abc.html
    """

    @property
    # @abstractmethod #!!! uncomment when I no longer assume the value
    def io_type(self):
        """
        Should I do StructureTransformation, LatticeTransformation, and
        SiteTransformation subclasses? Here, io_type could be either
        'structure', 'lattice', or 'site' as to indicate the input.
        """
        #!!! For now, I have all transformations as StructureTransformation
        #!!! so I assume io_type here. This may change in the future.
        return "structure"

    @property
    @abstractmethod
    def io_scale(self):
        """
        The input/output type for the class.
        This should be one of the following choices:
            - one_to_one
            - one_to_many
            - many_to_one
            - many_to_many # I have no examples of this yet
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def use_multiprocessing(self):
        """
        Whether or not this transformation can be done in parallel.
        Simply set to True or False.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def ninput(self):
        """
        Number of inputs required. For example, some transformations require
        two structures to be input. If one_to_* we know ninput = 1.
        """
        raise NotImplementedError

    @abstractmethod
    def apply_transformation(self):
        """
        The code that carries out the
        """
        raise NotImplementedError

    def apply_from_database_and_selector(
        self,
        selector,
        datatable,
        select_kwargs: dict = {},
        max_attempts: int = 100,
        **kwargs,  # for apply_transformation_with_validation
    ):

        logging.info(f"Creating a transformed structure with {self.__class__.__name__}")
        logging.info(f"Parent(s) will be selected using {selector.__class__.__name__}")

        # grab parent structures using the selection method
        parent_ids, parent_structures = selector.select_from_datatable(
            nselect=self.ninput,
            datatable=datatable,
            **select_kwargs,
        )

        # Until we get a new valid structure (or run out of attempts), keep trying
        # with our given source. Assume we don't have a valid structure until
        # proven otherwise
        new_structure = False
        attempt = 0
        while not new_structure and attempt <= max_attempts:
            # add an attempt
            attempt += 1

            # make a new structure
            new_structure = self.apply_transformation_with_validation(
                parent_structures,
                **kwargs,
            )

        # see if we got a structure or if we hit the max attempts and there's
        # a serious problem!
        if not new_structure:
            logging.warning(
                f"Failed to create a structure after {max_attempts} different"
                "parent combinations. Giving up."
            )
            return False

        # Otherwise we were successful
        logging.info("Creation Successful.")

        return parent_ids, new_structure

    def apply_transformation_with_validation(
        self,
        structures,  # either a structure or list of structures. Depends on ninput.
        validators=[],
        max_attempts=100,
    ):

        # Until we get a new valid structure (or run out of attempts), keep trying
        # with our given source. Assume we don't have a valid structure until
        # proven otherwise
        new_structure = False
        attempt = 0
        while not new_structure and attempt <= max_attempts:
            # add an attempt
            attempt += 1

            # make a new structure
            new_structure = self.apply_transformation(structures)

            # check to see if the structure passes all validation checks.
            if new_structure:

                for validator in validators:
                    is_valid = validator.check_structure(new_structure)

                    if not is_valid:
                        # if it is not unique, we can throw away the structure and
                        # try the loop again.
                        logging.debug(
                            "Generated structure failed validation by "
                            f"{validator.__class__.__name__}. Trying again."
                        )
                        new_structure = None

                        # one failed validation is enough to stop. There is no
                        # need to test the other validation methods.
                        break

        # see if we got a structure or if we hit the max attempts and there's
        # a serious problem!
        if not new_structure:
            logging.warning(
                "Failed to create a structure from input after "
                f"{max_attempts} attempts"
            )
            return False

        # return the structure and its parents
        return new_structure



# -*- coding: utf-8 -*-

from simmate.toolkit.structure_prediction.evolution.workflows.all import (
    StructurePrediction__Python__FixedComposition,
    StructurePrediction__Python__NewIndividual,
)



# -*- coding: utf-8 -*-

from .customized import Customized__Vasp__UserConfig
from .diffusion.all import (
    Diffusion__Vasp__NebAllPathsMit,
    Diffusion__Vasp__NebFromEndpointsMit,
    Diffusion__Vasp__NebFromImagesMit,
    Diffusion__Vasp__NebFromImagesMvlCi,
    Diffusion__Vasp__NebSinglePathMit,
)
from .dynamics.all import (
    Dynamics__Vasp__Matproj,
    Dynamics__Vasp__Mit,
    Dynamics__Vasp__MvlNpt,
)
from .electronic_structure.all import (
    ElectronicStructure__Vasp__MatprojFull,
    ElectronicStructure__Vasp__MatprojHseFull,
)
from .population_analysis.all import (
    PopulationAnalysis__Vasp__BadelfMatproj,
    PopulationAnalysis__Vasp__BaderMatproj,
    PopulationAnalysis__Vasp__ElfMatproj,
)
from .relaxation.all import (
    Relaxation__Vasp__Matproj,
    Relaxation__Vasp__MatprojHse,
    Relaxation__Vasp__MatprojMetal,
    Relaxation__Vasp__MatprojScan,
    Relaxation__Vasp__Mit,
    Relaxation__Vasp__MvlGrainboundary,
    Relaxation__Vasp__MvlNebEndpoint,
    Relaxation__Vasp__MvlSlab,
    Relaxation__Vasp__Quality00,
    Relaxation__Vasp__Quality01,
    Relaxation__Vasp__Quality02,
    Relaxation__Vasp__Quality03,
    Relaxation__Vasp__Quality04,
    Relaxation__Vasp__Staged,
)
from .static_energy.all import (
    StaticEnergy__Vasp__Matproj,
    StaticEnergy__Vasp__MatprojHse,
    StaticEnergy__Vasp__MatprojScan,
    StaticEnergy__Vasp__Mit,
    StaticEnergy__Vasp__MvlNebEndpoint,
    StaticEnergy__Vasp__Quality04,
)



# -*- coding: utf-8 -*-

"""
This moduel is a work in progress and therefore unused at the moment. 

This is an attempt to add coloring to logging.
"""

import logging

# All the code below is just to add coloring to our logger. If coloring isn't
# needed, the same result can be achieved with the following:
# logging.basicConfig(
#     format="SIMMATE-%(levelname)-s | %(asctime)-s | %(message)s",
#     level=logging.INFO,
#     datefmt="%Y-%m-%d %H:%M:%S",
# )


class CustomFormatter(logging.Formatter):
    """
    Logging colored formatter, adapted from the following sources:
    https://stackoverflow.com/a/56944256/3638629
    https://alexandra-zaharia.github.io/posts/make-your-own-custom-color-formatter-with-python-logging/
    """

    # These colors are ANSI escape codes
    # I don't fully understand them, but I used the following script to print
    # out all of the options
    # https://www.lihaoyi.com/post/BuildyourownCommandLinewithANSIescapecodes.html#256-colors
    #
    # import sys
    # for i in range(0, 16):
    #     for j in range(0, 16):
    #         code = str(i * 16 + j)
    #         sys.stdout.write(u"\u001b[38;5;" + code + "m " + code.ljust(4))
    #     print(u"\u001b[0m")
    #
    # The number corresponds to the final part of the code below (I think)

    debug_color = "\u001b[38;5;13m"  # purple
    info_color = "\u001b[38;5;15m"  # white
    warning_color = "\u001b[38;5;11m"  # yellow
    error_color = "\u001b[38;5;9m"  # red
    critical_color = "\u001b[38;5;9m"  # red
    reset = "\x1b[0m"

    datefmt = "%Y-%m-%d %H:%M:%S"

    def __init__(self, fmt):
        super().__init__()
        self.fmt = fmt
        self.FORMATS = {
            logging.DEBUG: self.debug_color + self.fmt + self.reset,
            logging.INFO: self.info_color + self.fmt + self.reset,
            logging.WARNING: self.warning_color + self.fmt + self.reset,
            logging.ERROR: self.error_color + self.fmt + self.reset,
            logging.CRITICAL: self.critical_color + self.fmt + self.reset,
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt, datefmt=self.datefmt)
        return formatter.format(record)


# Create custom logger logging all five levels
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)

# Define format for logs
fmt = "[SIMMATE-%(levelname)-s | %(asctime)-s] %(message)s"

# Create stdout handler for logging to the console (logs all five levels)
stdout_handler = logging.StreamHandler()
stdout_handler.setLevel(logging.DEBUG)
stdout_handler.setFormatter(CustomFormatter(fmt))

# Add both handlers to the logger
logger.addHandler(stdout_handler)

# Example logging and also for testing.
# logger.debug("this is a debug message")
# logger.info("this is a info message")
# logger.warning("this is a warning message")
# logger.error("this is a error message")
# logger.critical("this is a critical message")



# -*- coding: utf-8 -*-

import logging
import time
from pathlib import Path

from simmate.database.workflow_results import EvolutionarySearch as SearchDatatable
from simmate.toolkit import Composition
from simmate.workflow_engine import Workflow

# TODO
# StructurePrediction__Python__VariableTernaryComposition
#   --> calls VariableBinaryComposition strategically
#   --> might call FixedCompositionVariableNsites strategically too
# StructurePrediction__Python__VariableBinaryComposition
#   --> calls FixedCompositionVariableNsites strategically
# StructurePrediction__Python__FixedCompositionVariableNsites
#   --> calls FixedComposition strategically


class StructurePrediction__Python__FixedComposition(Workflow):

    use_database = False
    # consider making a calculation table so that the first step of saving
    # Evolutionary Search is done automatically. It will also ease the
    # tracking of output files when many fixed compositions are submitted.

    @classmethod
    def run_config(
        cls,
        composition: str | Composition,
        subworkflow_name: str | Workflow = "relaxation.vasp.staged",
        subworkflow_kwargs: dict = {},
        fitness_field: str = "energy_per_atom",
        max_structures: int = 3000,
        limit_best_survival: int = 250,
        nfirst_generation: int = 20,
        nsteadystate: int = 40,
        singleshot_sources: list[str] = [],
        steadystate_sources: list[tuple[float, str]] = [
            (0.30, "RandomSymStructure"),
            (0.30, "from_ase.Heredity"),
            (0.10, "from_ase.SoftMutation"),
            (0.10, "from_ase.MirrorMutation"),
            (0.05, "from_ase.LatticeStrain"),
            (0.05, "from_ase.RotationalMutation"),
            (0.05, "from_ase.AtomicPermutation"),
            (0.05, "from_ase.CoordinatePerturbation"),
        ],
        selector_name: str = "TruncatedSelection",
        selector_kwargs: dict = {},
        validator_name: str = "PartialCrystalNNFingerprint",
        validator_kwargs: dict = {},
        tags: list[str] = None,
        sleep_step: int = 60,
        directory: Path = None,
        **kwargs,
    ):
        """
        Sets up the search engine and its settings.

        #### Parameters

        - `composition`:
            The composition to run the evolutionary search for. Note that the
            number of sites is fixed to what is set here. (Ca2N vs Ca4N2)

        - `workflow`:
            The workflow to run all individuals through. Note, the result_database
            of this workflow will be treated as the individuals in this search.
            The default is "relaxation.vasp.staged"

        - `workflow_command`:
            The command that will be passed to the workflow.run() method.

        - `max_structures`:
            The maximum number of individuals that will be calculated before
            stopping the search. The default is 3000.

        - `limit_best_survival`:
            The search is stopped when the best individual remains unbeaten for
            this number of new individuals. The default is 250.

        - `singleshot_sources`:
            TODO: This is not implemented yet

        - `nfirst_generation`:
            No mutations or "child" individuals will be carried out until this
            number of individuals have been calculated. The default is 20.

        - `nsteadystate`:
            The total number of individuals from steady-state sources that will
            be running/submitted at any given time. The default is 40.

        - `steadystate_sources`:
            A list of tuples where each tuple is (percent, source). The percent
            determines the number of steady stage calculations that will be
            running for this at any given time. For example, 0.25 means
            0.25*40=10 individuals will be running/submitted at all times. The
            source can be from either the toolkit.creator or toolkit.transformations
            modules. Don't change this default unless you know what you're doing!

        - `selector`:
            The defualt method to use for choosing the parent individual(s). The
            default is TruncatedSelection.
        """

        #######################################################################
        # TODO: Check if there is an existing search and grab it if so.
        # We can also update, add logs, compare... I need to decide what to do
        # here. The stop condition could also be checked to see if a search
        # even needs to be started.
        # if SearchDatatable.objects.filter(composition=composition.formula).exists():
        #     self.search_datatable = SearchDatatable.objects.filter(
        #         composition=composition.formula,
        #     ).get()
        # Grab the list of singleshot sources that have been ran before
        # and based off of that list, remove repeated sources
        # past_singleshot_sources = (
        #     self.search_datatable.sources.filter(is_singleshot=True)
        #     .values_list("name", flat=True)
        #     .all()
        # )
        # singleshot_sources = [
        #     source
        #     for source in singleshot_sources
        #     if source not in past_singleshot_sources
        # ]
        # BUG: what if I want to rerun any of these even though its been ran before?
        # An example would be rerunning substituitions when new structures are available
        #######################################################################

        logging.info(f"Setting up evolutionary search for {composition}")
        # TODO: Should I add logs to inform how the workflow will be ran? Or
        # leave it to the simmate_metadata.yaml file and database for this?
        # logging.info(f"Individuals will be evaulated using '{subworkflow_name}'")
        # logging.info(f"Parent selection for mutations will use '{selector_name}'")

        #  Add the search entry to the DB.
        search_datatable = SearchDatatable(
            composition=composition.formula,
            subworkflow_name=subworkflow_name,
            subworkflow_kwargs=subworkflow_kwargs,
            fitness_field=fitness_field,
            max_structures=max_structures,
            limit_best_survival=limit_best_survival,
            nfirst_generation=nfirst_generation,
            nsteadystate=nsteadystate,
            selector_name=selector_name,
            selector_kwargs=selector_kwargs,
            validator_name=validator_name,
            validator_kwargs=validator_kwargs,
            tags=tags,
            sleep_step=sleep_step,
        )
        search_datatable.save()

        if search_datatable.check_stop_condition():
            logging.info("Looks like this search was already ran by someone else!")
            return

        # See if the singleshot sources have been ran yet. For restarted calculations
        # this will likely not be needed (unless a new source was added). But for
        # new searches/compositions, this will submit all individuals from the
        # single shot sources before we even start the steady-state runs
        search_datatable._check_singleshot_sources(directory)

        # Initialize the steady state sources by saving their config information
        # to the database.
        search_datatable._init_steadystate_sources_to_db(steadystate_sources)

        logging.info("Finished setup")
        logging.info(f"Assigned this to EvolutionarySearch id={search_datatable.id}.")

        # this loop will go until I hit 'break' below
        while True:

            # Write the output summary if there is at least one structure completed
            if search_datatable.individuals_completed.count() >= 1:
                search_datatable.write_summary(directory)
            else:
                search_datatable.write_individuals_incomplete(directory)

            # TODO: maybe write summary files to csv...? This may be a mute
            # point because I expect we can follow along in the web UI in the future
            # To that end, I can add a "URL" property

            # Check the stop condition
            # If it is True, we can stop the calc.
            if search_datatable.check_stop_condition():
                # TODO: should I sleep / wait for other calculations and
                # check again? Then write summary one last time..?
                break  # break out of the while loop
            # Otherwise, keep going!

            # TODO: Go through triggered actions that would update the database
            # table -- e.g. the workflow to run, the validators, etc.
            # self._check_triggered_actions()

            # Go through the running workflows and see if we need to submit
            # new ones to meet our steadystate target(s)
            search_datatable._check_steadystate_workflows()

            # To save our database load, sleep until we run checks again.
            logging.info(
                f"Sleeping for {sleep_step} seconds before running checks again."
            )
            time.sleep(sleep_step)

        logging.info("Stopping the search (running calcs will be left to finish).")



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate workflow engine. All commands are 
accessible through the "simmate workflow-engine" command.
"""

import pathlib

import click


@click.group()
def utilities():
    """
    A group of commands for various simple tasks - such as file handling.
    """
    pass


@utilities.command()
@click.argument(
    "directory",
    default=pathlib.Path.cwd(),
    type=click.Path(exists=True, path_type=pathlib.Path),
)
@click.option(
    "--time_cutoff",
    "-t",
    default=3 * 7 * 24 * 60 * 60,  # equal to 3 weeks
    type=float,
    help=(
        "if the queue is empty, the time (in seconds) the worker should wait"
        " before checking the queue again"
    ),
)
def archive_old_runs(directory, time_cutoff):
    """
    This starts a Simmate Worker which will query the database for jobs to run.
    """

    from simmate.utilities import archive_old_runs

    archive_old_runs(directory, time_cutoff)


# explicitly list functions so that pdoc doesn't skip them
__all__ = [
    "archive_old_runs",
]
